<?php

namespace Drupal\couchbasedrupal\Cache;

abstract class CouchbaseBackendGeneric {

  /**
   * Wether to do reliable reads. "Slower" but consistent.
   *
   * @var boolean
   */
  protected $reliable = TRUE;

  /**
   * Prefix for all keys in this cache bin.
   *
   * Includes the site-specific prefix in $sitePrefix.
   *
   * @var string
   */
  protected $binPrefix;

  /**
   * Couchbase bucket
   *
   * @var \CouchbaseBucket
   */
  protected $bucket;

  /**
   * Query options.
   *
   * @var array
   */
  protected $options = array();

  /**
   * Name of the view for this binary in the Couchbase database
   *
   * @var string
   */
  protected $view;

  /**
   * Prepends the Wincache user variable prefix for this bin to a cache item ID.
   *
   * @param string $cid
   *   The cache item ID to prefix.
   *
   * @return string
   *   The storage key for the cache item ID.
   */
  protected function getBinKey($cid) {
    $key = $this->binPrefix . $cid;
    // If you get beyound 250 bytes the PHP driver
    // for couchbase hangs.
    if (mb_strlen($key, '8bit') >= 250) {
      $key = hash('sha256', $this->binPrefix . $cid);
    }
    return $key;
  }

  /**
   * Make sure that we create a view for each binary.
   *
   * Not used for anything yet...
   *
   * @throws \Exception
   */
  protected function EnsureViews() {

    
  }

  protected function EnsureViewIndexUpdated() {
  
  }

  /**
   * Retrieve an item from the bucket without throwing
   * an Exception if it does not exist.
   *
   * @param string|string[] $key
   * @return mixed|bool
   */
  protected function bucketGet($key) {
    try {
      return $this->bucket->get($key);
    }
    catch (\Exception $e) {
      if ($e->getCode() == 13) {
        return FALSE;
      }
      throw $e;
    }
  }

  /**
   * Remove an item from the bucket without
   * throwing and exceptino if it does not exist.
   *
   * @param string|string[] $key
   * @return mixed|bool
   */
  protected function bucketRemove($key) {
    try {
      if (!is_array($key)) {
        $key = array($key);
      }
      foreach ($key as $id) {
        return $this->bucket->remove($id);
      }
    }
    catch (\Exception $e) {
      if ($e->getCode() == 13) {
        return FALSE;
      }
      throw $e;
    }
    return TRUE;
  }

  /**
   * Upsert an intem in the bucket.
   *
   * @param string $id
   * @param mixed $val
   * @param array $options
   */
  protected function bucketUpsert($id, $val = NULL, $options = array()) {
    $this->bucket->upsert($id, $val, $options);
  }

  /**
   * Insert an item in a bucket. On failure or if the item already exists
   * returns FALSE.
   *
   * @param string $id
   * @param mixed $val
   * @param mixed $options
   * @return bool
   */
  protected function bucketInsert($id, $val = NULL, $options = array()) {
    try {
      $this->bucket->insert($id, $val, $options);
      return TRUE;
    }
    catch (\Exception $e) {
      $code = $e->getCode();
      // 12,13 are document already exists or document does not exist.
      if (in_array($code, array(12, 13))) {
        return FALSE;
      }
      throw $e;
    }
  }
}
