<?php

/**
 * @file
 * Contains \Drupal\couchbasedrupal\Cache\CouchbaseRawBackend.
 */

namespace Drupal\couchbasedrupal\Cache;

use Drupal\supercache\Cache\CacheRawBackendInterface;
use Drupal\Core\Cache\Cache;

/**
 * Stores cache items in Couchbase.
 */
class CouchbaseRawBackend extends CouchbaseBackendGeneric implements CacheRawBackendInterface {

  /**
   * Last time this cache was cleared.
   *
   * @var float
   */
  protected $lastClearTime = 0;

  /**
   * Get or Set the last clear time.
   *
   * @param float $time
   */
  protected function LastClear($time = NULL) {
    $key = 'BinaryLastCleared';
    if (empty($time)) {
      if ($cache = $this->bucketGet($key)) {
        $this->lastClearTime = $cache->value;
      }
      $this->lastClearTime = 0;
    }
    else {
      $this->bucketUpsert($key, $time);
      $this->lastClearTime = $time;
    }
  }

  /**
   * Constructs a new CouchbaseBackend instance.
   *
   * @param string $bin
   *   The name of the cache bin.
   * @param string $site_prefix
   *   The prefix to use for all keys in the storage that belong to this site.
   * @param \Drupal\Core\Cache\CacheTagsChecksumInterface $checksum_provider
   *   The cache tags checksum provider.
   */
  public function __construct($binPrefix, \CouchbaseBucket $bucket) {
    $this->binPrefix = $binPrefix;
    $this->bucket = $bucket;
    $this->view = $this->binPrefix;
    $this->options = array();
    $this->LastClear();
    $this->EnsureViews();
  }

  /**
   * {@inheritdoc}
   */
  public function get($cid) {
    $cids = array($cid);
    $cache = $this->getMultiple($cids);
    return reset($cache);
  }

  /**
   * {@inheritdoc}
   */
  public function getMultiple(&$cids) {
    // Translate the requested cache item IDs to Wincache keys.
    $map = array();
    foreach ($cids as $cid) {
      $map[$this->getBinKey($cid)] = $cid;
    }

    $result = $this->bucket->get(array_keys($map));
    $cache = array();
    foreach ($result as $key => $item) {
      $item = $this->prepareItem($item->value);
      if ($item) {
        $cache[$map[$key]] = $item;
      }
    }
    unset($result);

    $cids = array_diff($cids, array_keys($cache));
    return $cache;
  }

  /**
   * Returns all cached items, optionally limited by a cache ID prefix.
   *
   *
   * @param string $prefix
   *   (optional) A cache ID prefix to limit the result to.
   *
   * @return string[]
   *   An array of Keys for this binary.
   */
  public function getAll($prefix = '') {
    $query = \CouchbaseViewQuery::from($this->view, 'keys'); //->custom(array('startkey' => "%22$prefix%22"));
    $rows = $this->bucket->query($query)['rows'];
    return array_column($rows, 'value');
  }

  /**
   * Returns all cached items for this binary.
   *
   * @return string[]
   */
  protected function getAllBinary() {
    throw new \Exception("Not implemented.");
  }

  /**
   * Prepares a cached item.
   *
   * Checks that the item is either permanent or did not expire.
   *
   * @param \stdClass $cache
   *   An item loaded from cache_get() or cache_get_multiple().
   * @param bool $allow_invalid
   *   If TRUE, a cache item may be returned even if it is expired or has been
   *   invalidated. See ::get().
   *
   * @return mixed
   *   The cache item or FALSE if the item expired.
   */
  protected function prepareItem($cache) {
    if (!isset($cache->data)) {
      return FALSE;
    }
    else {
      return (object) ['data' => $cache];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function set($cid, $data, $expire = CacheBackendInterface::CACHE_PERMANENT) {
    // If you don't pass an INT couchbase crashes.
    $expire = (int) $expire;
    // Prepare the item.
    if ($expire == CacheRawBackendInterface::CACHE_PERMANENT) {
      return $this->bucketUpsert($this->getBinKey($cid), $data);
    }
    else {
      return $this->bucketUpsert($this->getBinKey($cid), $data, array('expiry' => $expire));
    }
  }

  /**
   * Like set() but will return false (and fail) if the document already exists.
   */
  public function add($cid, $data, $expire = CacheBackendInterface::CACHE_PERMANENT) {
    // If you don't pass an INT couchbase crashes.
    $expire = (int) $expire;
    // Prepare the item.
    if ($expire == CacheRawBackendInterface::CACHE_PERMANENT) {
      return $this->bucketInsert($this->getBinKey($cid), $data);
    }
    else {
      return $this->bucketInsert($this->getBinKey($cid), $data, array('expiry' => $expire));
    }
  }

  /**
   * {@inheritdoc}
   */
  public function setMultiple(array $items = array()) {
    foreach ($items as $cid => $item) {
      $this->set($cid, $item['data'], isset($item['expire']) ? $item['expire'] : CacheRawBackendInterface::CACHE_PERMANENT);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function delete($cid) {
    $this->bucketRemove($this->getBinKey($cid));
  }

  /**
   * {@inheritdoc}
   */
  public function deleteMultiple(array $cids) {
    $keys = array_map(array($this, 'getBinKey'), $cids);
    foreach ($keys as $key) {
      $this->bucketRemove($key);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function deleteAll() {
    $this->LastClear(microtime(TRUE));
  }

  public function counter($cid, $increment, $default = 0, $expire = CacheRawBackendInterface::CACHE_PERMANENT) {
    $this->counterMultiple([$cid], $increment, $default, $expire);
  }

  public function counterMultiple(array $cids, $increment, $default = 0, $expire = CacheRawBackendInterface::CACHE_PERMANENT) {
    $keys = array_map(array($this, 'getBinKey'), $cids);
    $this->bucket->counter($keys, 1, array('initial' => $default));
  }

  /**
   * {@inheritdoc}
   */
  public function removeBin() {
    $this->bucketRemove($this->getAllBinary());
  }

  /**
   * {@inheritdoc}
   */
  public function invalidate($cid) {
    $this->invalidateMultiple(array($cid));
  }

  /**
   * {@inheritdoc}
   */
  public function invalidateMultiple(array $cids) {
    foreach ($this->getMultiple($cids) as $cache) {
      $this->bucket->touch($this->getBinKey($cache->cid), REQUEST_TIME - 1);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function touch($cid, $expire = CacheRawBackendInterface::CACHE_PERMANENT) {
    $this->bucket->touch($this->getBinKey($cid), (int) $expire);
  }

  /**
   * {@inheritdoc}
   */
  public function touchMultiple(array $cids, $expire = CacheRawBackendInterface::CACHE_PERMANENT) {
    foreach ($cids as $cid) {
      $this->touch($cid, $expire);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function invalidateAll() {
    foreach ($this->getAll() as $cid) {
      $this->bucket->touch($this->getBinKey($cid), REQUEST_TIME - 1);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function garbageCollection() {
    // Nothing to do here. 
    // Couchbase automatically handles expirations.
  }

}
