<?php

/**
 * @file
 * Contains \Drupal\couchbasedrupal\Cache\CouchbaseTagsChecksum.
 */

namespace Drupal\couchbasedrupal\Cache;

use Drupal\Core\Database\Connection;
use Drupal\Core\Database\SchemaObjectExistsException;
use Drupal\couchbasedrupal\CouchbaseManager;
use Drupal\Core\Cache\CacheTagsChecksumInterface;
use Drupal\Core\Cache\CacheTagsInvalidatorInterface;

/**
 * Cache tags invalidations checksum implementation that uses the database.
 *
 * Only use this if running on a single web head setup!
 *
 */
class CouchbaseTagsChecksum implements CacheTagsChecksumInterface, CacheTagsInvalidatorInterface {

  /**
   * Bucket.
   *
   * @var \CouchbaseBucket
   */
  protected $bucket;

  /**
   * Couchbase manager.
   *
   * @var CouchbaseManager
   */
  protected $couchbaseManager;

  /**
   * Contains already loaded cache invalidations from the database.
   *
   * @var array
   */
  protected $tagCache = array();

  /**
   * A list of tags that have already been invalidated in this request.
   *
   * Used to prevent the invalidation of the same cache tag multiple times.
   *
   * @var string[]
   */
  protected $invalidatedTags = array();

  /**
   * Retrieve the active bucket.
   * 
   * @return \CouchbaseBucket
   */
  protected function getBucket() {
    if (empty($this->bucket)) {
      $this->bucket = $this->couchbaseManager->getBucket();
    }
    return $this->bucket;
  }

  /**
   * Constructs a CouchbaseTagsChecksum object.
   *
   */
  public function __construct(CouchbaseManager $manager = NULL) {
    $this->couchbaseManager = $manager;
  }

  /**
   * {@inheritdoc}
   */
  public function invalidateTags(array $tags) {
    $keys_to_increment = array();
    foreach ($tags as $tag) {
      // Only invalidate tags once per request unless they are written again.
      if (isset($this->invalidatedTags[$tag])) {
        continue;
      }
      $keys_to_increment[] = $tag;
      $this->invalidatedTags[$tag] = TRUE;
      unset($this->tagCache[$tag]);
    }
    // We can increment all of them on the storage at once.
    $keys_to_increment = preg_filter('/^/', 'cache_tag::', $keys_to_increment);
    $this->getBucket()->counter($keys_to_increment, 1, array('initial' => 0));
  }

  /**
   * {@inheritdoc}
   */
  public function getCurrentChecksum(array $tags) {
    // Remove tags that were already invalidated during this request from the
    // static caches so that another invalidation can occur later in the same
    // request. Without that, written cache items would not be invalidated
    // correctly.
    foreach ($tags as $tag) {
      unset($this->invalidatedTags[$tag]);
    }
    return $this->calculateChecksum($tags);
  }

  /**
   * {@inheritdoc}
   */
  public function isValid($checksum, array $tags) {
    return $checksum == $this->calculateChecksum($tags);
  }

  /**
   * Calculates the current checksum for a given set of tags.
   *
   * @param array $tags
   *   The array of tags to calculate the checksum for.
   *
   * @return int
   *   The calculated checksum.
   */
  protected function calculateChecksum(array $tags) {
    $missing_tags = array_diff($tags, array_keys($this->tagCache));

    if ($missing_tags) {
      $keys = preg_filter('/^/', 'cache_tag::', $tags);
      $stored_tags = $this->getBucket()->get($keys);
      foreach ($stored_tags as $tag) {
        if (empty($tag->error)) {
          $this->tagCache[] = $tag->value;
        }
      }
      // Fill static cache with empty objects for tags not found in the database.
      $this->tagCache += array_fill_keys(array_diff($missing_tags, array_keys($stored_tags)), 0);
    }

    $real_tags = array_intersect_key($this->tagCache, array_flip($tags));
    return array_sum($real_tags);
  }

  /**
   * {@inheritdoc}
   */
  public function reset() {
    $this->tagCache = array();
    $this->invalidatedTags = array();
  }
}
