<?php

namespace Drupal\couchbasedrupal;

use Drupal\Core\Site\Settings;

use Psr\Log\LoggerInterface;

/**
 * This is a service used to manage the connections
 * to the different couchbase servers.
 */
class CouchbaseManager {

  /**
   * Couchbase server settings.
   *
   * @var array
   */
  var $settings;

  /**
   * Logger service.
   *
   * @var LoggerInterface
   */
  var $logger;

  /**
   * Settings
   *
   * @param Settings $settings
   */
  public function __construct(Settings $settings, LoggerInterface $logger) {
    // Add a default for a local couchbase server.
    $this->settings = ['servers' =>
      ['default' => ['uri' => 'couchbase://127.0.0.1'] ]
    ];
    // Merge or override with custom settings.
    array_merge($this->settings, $settings::get('couchbasedrupal', array()));
    $this->logger = $logger;
  }

  /**
   * List of instanced clusters.
   *
   * @var \CouchbaseCluster[]
   */
  protected $clusters = array();

  /**
   * List of instanced buckets.
   *
   * @var \CouchbaseBucket[]
   */
  protected $buckets = array();

  /**
   * Retrieve a connection to the couchbase server.
   *
   * @param string $name
   * @return bool|\CouchbaseCluster
   */
  public function getCluster($name = 'default') {

    // For this service to work we need the CouchbaseCluster class.
    if (!class_exists(\CouchbaseCluster::class)) {
      $this->logger->notice("CouchbaseCluster class does not exist. Make sure the couchbase extension is up and running.");
      return FALSE;
    }

    if (!isset($this->settings['servers'][$name]['uri'])) {
      $this->logger->notice("Couchbase server settings for $name not found");
      return FALSE;
    }

    $uri = $this->settings['servers']['default']['uri'];

    try {
      $server = $this->ParseUri($uri);
    }
    catch (\Exception $e) {
      $this->logger->notice("Could not parse uri for couchbase server: $uri");
      return FALSE;
    }

    if (!isset($this->clusters[$uri])) {
      if (isset($server['username']) && isset($server['password'])) {
        $this->clusters[$uri] = new \CouchbaseCluster($server['scheme'] . '://' . $server['host'], $server['username'], $server['password']);
      }
      else {
        $this->clusters[$uri] = new \CouchbaseCluster($server['scheme'] . '://' . $server['host']);
      }
    }

    return $this->clusters[$uri];
  }

  /**
   * Retrieve a couchbase bucket.
   *
   * @param string $cluster
   * @param string $bucket
   * @return \CouchbaseBucket
   */
  public function getBucket($cluster = 'default', $bucket = 'default') {
    $cluster = $this->getCluster($cluster);
    if (!isset($this->buckets[$bucket])) {
      $this->buckets[$bucket] = $cluster->openBucket($bucket);
    }
    return  $this->buckets[$bucket];
  }

  /**
   * Parse a couchbase connection URI.
   *
   * @param string $uri
   */
  protected function ParseUri($url) {
    $info = parse_url($url);
    if (!isset($info['scheme'], $info['host'])) {
      throw new \InvalidArgumentException('Minimum requirement: driver://host');
    }
    $info += array(
      'user' => '',
      'pass' => '',
      'fragment' => '',
    );

    $database = array(
      'scheme' => $info['scheme'],
      'username' => $info['user'],
      'password' => $info['pass'],
      'host' => $info['host'],
    );

    if (isset($info['port'])) {
      $database['port'] = $info['port'];
    }

    return $database;
  }
}
